import {FC, FormEventHandler, useEffect, useRef, useState} from 'react';
import {models} from '@tonality/library-common';
import Uuid from '@tonality/library-uuid';
import {ALLOWED_DURATIONS, toRawDuration} from '@tonality/library-song-utils';
import { LeftSidebarWithMenu } from '@theoryofnekomata/viewfinder'
import styled from 'styled-components';
import MusicalKeyboard from '@theoryofnekomata/react-musical-keyboard';
import TextInput from '../../../molecules/forms/TextInput';
import TextArea from '../../../molecules/forms/TextArea';
import ActionButton from '../../../molecules/forms/ActionButton';
import ToggleButton from '../../../molecules/forms/ToggleButton';
import NumericInput from '../../../molecules/forms/NumericInput';
import Timer from '../../../molecules/timer/Timer';

const KeyboardContainer = styled(LeftSidebarWithMenu.ContentContainer)({
	'--color-natural-key': 'var(--color-bg, white)',
	'--color-accidental-key': 'var(--color-fg, black)',
	maxWidth: 'none',
	marginTop: '1rem',
	marginBottom: '1rem',
	padding: 0,
	overflow: 'auto',
	'@media (min-width: 1080px)': {
		maxWidth: 'calc(var(--config-base-width,360) * 2)',
		padding: '0 1rem',
	},
})

const KeyboardSizer = styled('div')({
	width: 1500,
	'@media (min-width: 1080px)': {
		width: '100%',
	},
})

const Form = styled('form')({});

const FormContents = styled('div')({
	display: 'grid',
	gap: '1rem',
})

const Primary = styled('div')({
	display: 'grid',
	gridTemplateColumns: 'auto 6rem',
	gap: '1rem',
})

const Toolbar = styled('div')({
	display: 'grid',
	gap: '1rem',
	fontSize: '1.5rem',
	'@media (min-width: 720px)': {
		gridTemplateColumns: '2fr 1fr',
	},
})

const DurationSelector = styled('div')({
	display: 'grid',
	gridTemplateColumns: 'repeat(6, 1fr)',
	gap: '0.25rem',
})

const OtherTools = styled('div')({
	display: 'grid',
	gridTemplateColumns: 'repeat(3, 1fr)',
	gap: '1rem',
})

const NOTE_GLYPHS = {
	1: '𝅝',
	2: '𝅗𝅥',
	4: '♩',
	8: '♪',
	16: '𝅘𝅥𝅯',
	32: '𝅘𝅥𝅰',
};

const REST_GLYPHS = {
	1: '𝄻',
	2: '𝄼',
	4: '𝄽',
	8: '𝄾',
	16: '𝄿',
	32: '𝅀',
};

type Props = {
	onSubmit?: FormEventHandler,
	action?: string,
	labels: Record<string, string>,
	defaultValues?: Partial<models.Ringtone>,
	updateTempo: ({ dataRef, setTempo, }) => (...args: unknown[]) => unknown,
	updateView: ({ setNoteGlyph, setRestGlyph, noteGlyphs, restGlyphs, }) => (...args: unknown[]) => unknown,
	addRest: ({ formRef, dataRef, }) => (...args: unknown[]) => unknown,
	addNote: ({ dataRef, formRef, }) => (...args: unknown[]) => unknown,
	togglePlayback: ({ formRef, setPlayTimestamp, setPlaying, }) => (...args: unknown[]) => unknown,
	updateSong: ({ formRef, }) => (...args: unknown[]) => unknown,
	play: ({ dataRef, playing, setPlaying, setNotes, playTimestamp, currentTime, }) => unknown,
}

const CreateRingtoneForm: FC<Props> = ({
	onSubmit,
	action,
	labels,
	defaultValues = {},
	updateTempo,
	updateView,
	addRest,
	addNote,
	togglePlayback,
	updateSong,
	play,
}) => {
	const [hydrated, setHydrated] = useState(false);
	const [tempo, setTempo] = useState(defaultValues.tempo || 120);
	const [currentTime, setCurrentTime] = useState(0)
	const [noteGlyph, setNoteGlyph] = useState(NOTE_GLYPHS[4]);
	const [restGlyph, setRestGlyph] = useState(REST_GLYPHS[4]);
	const [playTimestamp, setPlayTimestamp] = useState(0)
	const [playing, setPlaying] = useState(false)
	const [notes, setNotes] = useState([])
	const dataRef = useRef<HTMLTextAreaElement>(null);
	const formRef = useRef<HTMLFormElement>(null);

	const handleTick = (now) => {
		setCurrentTime(now)
	}

	useEffect(() => {
		if (typeof play === 'function') {
			play({dataRef, playing, currentTime, setNotes, setPlaying, playTimestamp,})
		}
	}, [playTimestamp, playing, currentTime, setNotes, setPlaying, dataRef])

	useEffect(() => {
		setHydrated(true)
	}, [])

	// @ts-ignore
	return (
		<>
			<Form
				onSubmit={onSubmit}
				method="post"
				action={action}
				aria-label={labels['form']}
				ref={formRef}
			>
				<LeftSidebarWithMenu.ContentContainer>
					<FormContents>
						<input
							type="hidden"
							name="id"
							defaultValue={defaultValues.id ? Uuid.parse(defaultValues.id.toString()).toString() : undefined}
						/>
						<input
							type="hidden"
							name="composerUserSub"
							defaultValue={defaultValues.composerUserSub}
						/>
						<Primary>
							<TextInput
								label={labels['name'] || 'Name'}
								name="name"
								block
								defaultValue={defaultValues.name}
							/>
							<NumericInput
								label={labels['tempo'] || 'Tempo'}
								name="tempo"
								block
								defaultValue={defaultValues.tempo || 120}
								onChange={typeof updateTempo === 'function' ? updateTempo({ dataRef, setTempo, }) : undefined}
								disabled={playing}
							/>
						</Primary>
						{
							hydrated
							&& (
								<Toolbar>
									<DurationSelector>
										{
											ALLOWED_DURATIONS.map(d => (
												<ToggleButton
													key={d}
													name="duration"
													value={d.toString()}
													type="radio"
													onChange={typeof updateView === 'function' ? updateView({ setNoteGlyph, setRestGlyph, noteGlyphs: NOTE_GLYPHS, restGlyphs: REST_GLYPHS, }) : undefined}
													disabled={playing}
													defaultChecked={d === 4}
												>
													{NOTE_GLYPHS[d]}
												</ToggleButton>
											))
										}
									</DurationSelector>
									<OtherTools>
										<ToggleButton
											name="dotted"
											type="checkbox"
											disabled={playing}
										>
											{noteGlyph + ' + .'}
										</ToggleButton>
										<ActionButton
											onClick={typeof addRest === 'function' ? addRest({ formRef, dataRef, }) : undefined}
											disabled={playing}
										>
											{restGlyph}
										</ActionButton>
										<ActionButton
											onClick={typeof togglePlayback === 'function' ? togglePlayback({ formRef, setPlayTimestamp, setPlaying, }) : undefined}
										>
											{playing ? '■' : '▶'}
										</ActionButton>
									</OtherTools>
								</Toolbar>
							)
						}
						<TextArea
							label={labels['data'] || 'Data'}
							name="data"
							block
							defaultValue={defaultValues.data}
							readOnly={playing}
							ref={dataRef}
							onBlur={typeof updateSong === 'function' ? updateSong({ formRef, }) : undefined}
							style={{
								height: '6rem',
							}}
						/>
					</FormContents>
				</LeftSidebarWithMenu.ContentContainer>
				{
					hydrated
					&& (
						<KeyboardContainer>
							<KeyboardSizer>
								<MusicalKeyboard
									startKey={60}
									endKey={60 + (12 * 4) - 1}
									onChange={!playing && typeof addNote === 'function' ? addNote({ dataRef, formRef, }) : undefined}
									keyLabels={k => {
										if (k % 12 === 0) {
											return (Math.floor(k / 12) - 1).toString();
										}
										return '';
									}}
									keysOn={
										playing
											? notes
											: []
									}
								/>
							</KeyboardSizer>
						</KeyboardContainer>
					)
				}
				<LeftSidebarWithMenu.ContentContainer>
					<ActionButton
						type="submit"
						block
					>
						{labels['cta'] || 'Post'}
					</ActionButton>
				</LeftSidebarWithMenu.ContentContainer>
			</Form>
			<Timer
				onTick={handleTick}
				interval={toRawDuration(32, tempo)}
				resume={playing}
				start={playTimestamp}
			/>
		</>
	);
};

export default CreateRingtoneForm;
