import getFormValues from '@theoryofnekomata/formxtra';
import {
	Duration,
	parseSong,
	PITCH_CLASSES,
	Playable,
	SongData,
	stringifyNote,
	toRawDuration,
} from '@tonality/library-song-utils'
import SoundManager from '../../utils/sound/SoundManager'

interface PlayableWithOscillator extends Playable {
	played: boolean,
	stopped: boolean,
}

export default class ComposerClient {
	private songRef: SongData<PlayableWithOscillator>

	constructor(private readonly soundManagerRef: SoundManager) {}

	addNote = ({ dataRef, formRef, }) => (args) => {
		if (!(args.length > 0 && dataRef.current && formRef.current)) {
			return;
		}
		const values = getFormValues(formRef.current);
		const isDotted = Boolean(values['dotted'])
		const duration = Number(values['duration']) as Duration
		const tempo = Number(values['tempo'])
		const rawDuration = toRawDuration(duration, tempo)
		const formattedKeyData = stringifyNote({
			octave: Math.floor((args[0].key / 12) - 1),
			pitchClass: PITCH_CLASSES[args[0].key % 12],
			dotted: isDotted,
			duration,
		});
		const newSongData = `${dataRef.current.value} ${formattedKeyData}`.trim()
		dataRef.current.value = newSongData;
		this.songRef = parseSong(newSongData, Number(values['tempo']))
		this.soundManagerRef.start(args[0].key)
		setTimeout(() => {
			this.soundManagerRef.stop(args[0].key)
		}, rawDuration)
	};

	addRest = ({ formRef, dataRef, }) => () => {
		if (!(formRef.current && dataRef.current)) {
			return;
		}
		const values = getFormValues(formRef.current);
		const formattedKeyData = stringifyNote({
			dotted: Boolean(values['dotted']),
			duration: Number(values['duration']) as Duration,
		});
		const newSongData = `${dataRef.current.value} ${formattedKeyData}`.trim()
		dataRef.current.value = newSongData;
		this.songRef = parseSong(newSongData, Number(values['tempo']))
	};

	updateTempo = ({ dataRef, setTempo, }) => (e) => {
		if (!isNaN(e.target.valueAsNumber) && e.target.valueAsNumber > 0) {
			this.songRef = parseSong(dataRef.current.value, e.target.valueAsNumber)
			setTempo(e.target.valueAsNumber)
		}
	}

	updateSong = ({ formRef, }) => () => {
		const values = getFormValues(formRef.current);
		this.songRef = parseSong(values['data'], Number(values['tempo']))
	}

	updateView = ({ setNoteGlyph, setRestGlyph, noteGlyphs, restGlyphs, }) => (e) => {
		const duration = Number(e.target.value);
		setNoteGlyph(noteGlyphs[duration]);
		setRestGlyph(restGlyphs[duration]);
	}

	togglePlayback = ({ setPlayTimestamp, setPlaying, }) => () => {
		setPlayTimestamp(Date.now())
		setPlaying(p => {
			const newValue = !p
			this.songRef.playables.forEach(n => {
				if (!isNaN(n.key)) {
					if (!newValue && !n.stopped) {
						this.soundManagerRef.stop(n.key)
					}
				}
				n.played = newValue ? false : n.played
				n.stopped = newValue ? false : n.stopped
			})
			return newValue
		})
	}

	play = ({ dataRef, playing, currentTime, setNotes, setPlaying, playTimestamp, }) => {
		if (!playing) {
			return
		}

		const cursor = currentTime - 100 - playTimestamp
		if (cursor >= this.songRef.duration) {
			setPlaying(false)
			dataRef.current.setSelectionRange(null, null)
			this.soundManagerRef.stop(this.songRef.playables.slice(-1)[0].key)
		}

		this.songRef.playables.forEach((n) => {
			const toPlay = !n.played && n.start < cursor
			const toStop = !n.stopped && n.end < cursor
			if (toPlay) {
				if (!isNaN(n.key)) {
					this.soundManagerRef.start(n.key)
				}
				dataRef.current.select()
				dataRef.current.setSelectionRange(n.startIndex, n.endIndex)
				n.played = true
				setNotes(oldNotes => [...oldNotes, { velocity: 127, key: n.key }])
			}

			if (toStop) {
				if (!isNaN(n.key)) {
					this.soundManagerRef.stop(n.key)
				}
				n.stopped = true
				setNotes(oldNotes => oldNotes.filter(c => c.key !== n.key))
			}
		})
	}
}
