import {inject, singleton} from 'tsyringe'
import {models} from '@tonality/library-common'
import Uuid from '@tonality/library-uuid'
import {Controller, ControllerResponse, PaginatedResponse} from '../../utils/helpers'
import RingtoneService from './service'
import {
	InsufficientDataError,
	OwnershipError,
	RingtoneNotFoundError,
	UpdateDeletedRingtoneError,
} from './response';

type RingtoneController = Controller<{
	get: ControllerResponse<models.Ringtone>,
	browse: PaginatedResponse<models.Ringtone>,
	search: ControllerResponse<models.Ringtone>,
	create: ControllerResponse<models.Ringtone>,
	update: ControllerResponse<models.Ringtone>,
	softDelete: ControllerResponse<models.Ringtone>,
	undoDelete: ControllerResponse<models.Ringtone>,
	hardDelete: undefined,
}>

export default RingtoneController

const serializeRingtone = d => ({
	...d,
	id: d.id.toString(),
})

@singleton()
export class RingtoneControllerImpl implements RingtoneController {
	constructor(
		@inject('RingtoneService')
		private readonly ringtoneService: RingtoneService,
	) {}

	get = async (request: any, reply: any) => {
		try {
			const id = Uuid.parse(request.params['id'])
			const rawData = await this.ringtoneService.get(id, request.user?.sub)
			const data = serializeRingtone(rawData)
			if (data.deletedAt instanceof Date) {
				reply.raw.statusMessage = 'Ringtone Deleted Previously'
				reply.gone()
				return
			}
			reply.raw.statusMessage = 'Single Ringtone Retrieved'
			return {
				data,
			}
		} catch (err) {
			if (err instanceof RingtoneNotFoundError) {
				reply.raw.statusMessage = 'Ringtone Not Found'
				reply.notFound()
				return
			}
			reply.raw.statusMessage = 'Get Ringtone Error'
			reply.internalServerError(err.message)
		}
	}

	browse = async (request: any, reply: any) => {
		try {
			// TODO deserialize query
			const { 'skip': skipRaw, 'take': takeRaw } = request.query
			const skipNumber = Number(skipRaw)
			const takeNumber = Number(takeRaw)

			const skip = !isNaN(skipNumber) ? skipNumber : 0
			const take = !isNaN(takeNumber) ? takeNumber : 10
			const rawData = await this.ringtoneService.browse(skip, take, request.user?.sub)
			const data = rawData.data.map(serializeRingtone)

			reply.raw.statusMessage = 'Multiple Ringtones Retrieved'
			return {
				total: rawData.total,
				data,
				skip,
				take,
			}
		} catch (err) {
			reply.raw.statusMessage = 'Browse Ringtones Error'
			reply.internalServerError(err.message)
		}
	}

	search = async (request: any, reply: any) => {
		try {
			const { 'q': query } = request.query
			const rawData = await this.ringtoneService.search(query, request.user?.sub)
			const data = rawData.data.map(serializeRingtone)

			reply.raw.statusMessage = 'Search Results Retrieved'
			return {
				total: rawData.total,
				data,
			}
		} catch (err) {
			reply.raw.statusMessage = 'Search Error'
			reply.internalServerError(err.message)
		}
	}

	create = async (request: any, reply: any) => {
		// TODO parse and validate body

		try {
			const rawData = await this.ringtoneService.create(request.body, request.user?.sub)
			const data = serializeRingtone(rawData)
			reply.raw.statusMessage = 'Ringtone Created'
			reply.raw.statusCode = 201
			return {
				data,
			}
		} catch (err) {
			if (err instanceof InsufficientDataError) {
				reply.raw.statusMessage = 'Create Ringtone Insufficient Data Error'
				reply.badRequest(err.message)
				return
			}
			if (err instanceof OwnershipError) {
				reply.raw.statusMessage = 'Create Ringtone Ownership Error'
				reply.forbidden(err.message)
				return
			}
			reply.raw.statusMessage = 'Create Ringtone Error'
			reply.internalServerError(err.message)
		}
	}

	update = async (request: any, reply: any) => {
		try {
			// TODO validate data
			const id = Uuid.parse(request.params['id'])
			const rawData = await this.ringtoneService.update({
				...request.body,
				id,
			}, request.user?.sub)
			const data = serializeRingtone(rawData)
			reply.raw.statusMessage = 'Ringtone Updated'
			return {
				data,
			}
		} catch (err) {
			if (err instanceof RingtoneNotFoundError) {
				reply.raw.statusMessage = 'Ringtone Not Found'
				reply.notFound()
				return
			}
			if (err instanceof OwnershipError) {
				reply.raw.statusMessage = 'Update Ringtone Ownership Error'
				reply.forbidden(err.message)
				return
			}
			if (err instanceof UpdateDeletedRingtoneError) {
				reply.raw.statusMessage = 'Ringtone Deleted Previously'
				reply.gone()
				return
			}
			reply.raw.statusMessage = 'Update Ringtone Error'
			reply.internalServerError(err.message)
		}
	}

	softDelete = async (request: any, reply: any) => {
		try {
			// TODO validate data
			const id = Uuid.parse(request.params['id'])
			const rawData = await this.ringtoneService.softDelete(id, request.user?.sub)
			const data = serializeRingtone(rawData)
			reply.raw.statusMessage = 'Ringtone Soft-Deleted'
			return {
				data,
			}
		} catch (err) {
			if (err instanceof RingtoneNotFoundError) {
				reply.raw.statusMessage = 'Ringtone Not Found'
				reply.notFound()
				return
			}
			if (err instanceof OwnershipError) {
				reply.raw.statusMessage = 'Soft-Delete Ringtone Ownership Error'
				reply.forbidden(err.message)
				return
			}
			reply.raw.statusMessage = 'Soft-Delete Ringtone Error'
			reply.internalServerError(err.message)
		}
	}

	undoDelete = async (request: any, reply: any) => {
		try {
			// TODO validate data
			const id = Uuid.parse(request.params['id'])
			const rawData = await this.ringtoneService.undoDelete(id, request.user?.sub)
			const data = serializeRingtone(rawData)
			reply.raw.statusMessage = 'Ringtone Restored'
			return {
				data,
			}
		} catch (err) {
			if (err instanceof RingtoneNotFoundError) {
				reply.raw.statusMessage = 'Ringtone Not Found'
				reply.notFound()
				return
			}
			if (err instanceof OwnershipError) {
				reply.raw.statusMessage = 'Soft-Delete Ringtone Ownership Error'
				reply.forbidden(err.message)
				return
			}
			reply.raw.statusMessage = 'Restore Ringtone Error'
			reply.internalServerError(err.message)
		}
	}

	hardDelete = async (request: any, reply: any) => {
		try {
			// TODO validate data
			const id = Uuid.parse(request.params['id'])
			await this.ringtoneService.hardDelete(id, request.user?.sub)
			reply.status(204)
			reply.raw.statusMessage = 'Ringtone Hard-Deleted'
		} catch (err) {
			if (err instanceof RingtoneNotFoundError) {
				reply.raw.statusMessage = 'Ringtone Not Found'
				reply.notFound()
				return
			}
			if (err instanceof OwnershipError) {
				reply.raw.statusMessage = 'Delete Ringtone Ownership Error'
				reply.forbidden(err.message)
				return
			}
			reply.raw.statusMessage = 'Delete Ringtone Error'
			reply.internalServerError(err.message)
		}
	}
}
